! ----------------------------------------------------------------------
! This is part of the 'Solve Singular BVP' (SOLVSING) package developed
! by W.Auzinger, O.Koch, P.Kofler, E.Weinmueller,
! Department of Applied Mathematics and Numerical Analysis,
! Vienna University of Technology, Vienna, Austria.
! ----------------------------------------------------------------------
MODULE dynvect
! 'DYNamic expanding VECTor' (DYNVECT)
! Version 1.01, last modified 16-09-2000.
! --- Function and Algorithm ---
! This module provides dynamic vectors similar the Java implementation
! of the class 'java.util.Vector' (1.62 99/04/22), which is Copyright
! by Sun Microsystems, Inc., U.S.A. All rights reserved. But the
! functionality is less than in Java, only resetting, adding and copying
! to an array is supplied. As needed for RKPair, dynamic storage for a 
! vector and a real scalar is implemented.
! --- Compiling ---
! This module has to be compiled second, following the types module. 
! --- Bibliography ---
! M.Metcalf, J.Reid,
! 'Fortran 90 Explained', Oxford University Press (1990),
! L.Boynton, J.Payne, 'java.util.Vector.java', JDK1.22, Sun.
! ----------------------------------------------------------------------
! Constants
! ----------------------------------------------------------------------
   USE TYPES                 ! predefined data-types
   IMPLICIT NONE
   INTEGER(KIND = i4b), PARAMETER :: vsize_init = 100, &
                                     vsize_mult = 2
!        vsize_init ... initial size of the dynamic arrays
!        vsize_mult ... multiplier for the size if it has to grow

! ----------------------------------------------------------------------
! Global variables
! ----------------------------------------------------------------------
   REAL(KIND = dp), POINTER :: vectors(:,:), scalars(:)
!        vectors ... array for accumulating vectors
!        scalars ... array for accumulating scalars
   INTEGER(KIND = i4b) :: vsize, vsize_max
!        vsize ... size of the used fields of vectors and scalars
!        vsize_max ... size of complete array of vectors and scalars

   PRIVATE
   PUBLIC :: dv_init, dv_capacity, dv_size, dv_addelement, dv_toarray, &
             dv_removeallelements, dv_release

! ----------------------------------------------------------------------
! Subroutines
! ----------------------------------------------------------------------
CONTAINS

SUBROUTINE dv_init(dim)
! --- Function and Algorithm ---
! Set up the data for the dynamic vector to the specified dimension.
! This functionality is similar to Java's 'new java.util.Vector()'.
! --- Arguments ---
   IMPLICIT NONE
   INTEGER(KIND = i4b), INTENT(in) :: dim
!        dim ... dimension of the vectors to be stored in vectors.
! --- Code ---
   IF (ASSOCIATED(vectors)) DEALLOCATE(vectors) ! drop old values
   IF (ASSOCIATED(scalars)) DEALLOCATE(scalars) ! drop old values
   vsize_max = vsize_init
   vsize = 0             ! there are no values stored
   ALLOCATE(vectors(0:vsize_max-1,dim))
   ALLOCATE(scalars(0:vsize_max-1))
   vectors = 0.0_dp
   scalars = 0.0_dp
   RETURN
END SUBROUTINE dv_init

FUNCTION dv_capacity()
! --- Function and Algorithm ---
! Returns the current capacity of this vector, i.e. the length of its
! internal data arary.
! --- Arguments ---
   IMPLICIT NONE
   INTEGER(KIND = i4b) :: dv_capacity
! --- Code ---
   dv_capacity = vsize_max
   RETURN
END FUNCTION dv_capacity

FUNCTION dv_size()
! --- Function and Algorithm ---
! Returns the number of components in this vector.
! --- Arguments ---
   IMPLICIT NONE
   INTEGER(KIND = i4b) :: dv_size
! --- Code ---
   dv_size = vsize
   RETURN
END FUNCTION dv_size

SUBROUTINE dv_addelement(x,y)
! --- Function and Algorithm ---
! Adds a scalar and a vector-value to the end of this dynamic vector
! increasing its size by one.. If the array is too small to hold the
! next value, it is incresed.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: x, y(:)
!        x ...  the scalar to store next in scalars
!        y ... the value to store next in vectors
! --- Local variables ---
   REAL(KIND = dp), POINTER :: temps(:), tempv(:,:)
!        temp ... temporary for holding data while doubling the arrays
   INTEGER(KIND = i4b) :: new_size
!        new_size ... new (incresed) size of the dynamic vector
! --- Code ---
   IF (vsize >= vsize_max) THEN ! no space is left in dynamic vector
      new_size = vsize_max*vsize_mult

      ! increse the space of vectors
      ALLOCATE(tempv(0:vsize_max-1,SIZE(y))) ! create temp storage
      tempv = vectors           ! copy values to temporary storage
      DEALLOCATE(vectors)       ! drop small storage
      ALLOCATE(vectors(0:new_size-1,SIZE(y))) ! allocate more
      vectors(0:vsize_max-1,:) = tempv ! copy values from temp storage
      vectors(vsize_max:new_size-1,:) = 0.0_dp
      DEALLOCATE(tempv)      ! drop temporary storage

      ! increse the space of scalars
      ALLOCATE(temps(0:vsize_max-1)) ! create temp storage
      temps = scalars        ! copy values to temporary storage
      DEALLOCATE(scalars)    ! drop small storage
      ALLOCATE(scalars(0:new_size-1)) ! allocate more
      scalars(0:vsize_max-1) = temps ! copy values from temp storage
      scalars(vsize_max:new_size-1) = 0.0_dp
      DEALLOCATE(temps)      ! drop temporary storage

      vsize_max = new_size   ! we allocated more storage
   END IF
   vectors(vsize,:) = y
   scalars(vsize) = x
   vsize = vsize + 1         ! we added a new element to the next column
   RETURN
END SUBROUTINE dv_addelement

SUBROUTINE dv_toarray(grid,solution)
! --- Function and Algorithm ---
! Returns the arrays containing scalars and vectors all of the elements
! in this dynamic Vector in the correct order and in its exact size.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(out) :: grid(:)
   REAL(KIND = dp), INTENT(out) :: solution(:,:)
!        grid ... the field where the scalars are put
!        solution ... the field where the vectors are put
! --- Code ---
   ! no range check is done here !
   grid = scalars(0:vsize-1)
   solution = vectors(0:vsize-1,:)
   RETURN
END SUBROUTINE dv_toarray

SUBROUTINE dv_removeallelements()
! --- Function and Algorithm ---
! Removes all of the elements from this Vector. The Vector will be
! empty after this call returns
! --- Arguments ---
   IMPLICIT NONE
! --- Code ---
   vsize = 0                 ! there are no values stored
   vectors = 0.0_dp
   scalars = 0.0_dp
   RETURN
END SUBROUTINE dv_removeallelements

SUBROUTINE dv_release()
! --- Function and Algorithm ---
! return the used allocated arrays to the system.
! --- Arguments ---
   IMPLICIT NONE
! --- Code ---
   IF (ASSOCIATED(vectors)) DEALLOCATE(vectors) ! drop old values
   IF (ASSOCIATED(scalars)) DEALLOCATE(scalars) ! drop old values
   vsize_max = 0
   vsize = 0                 
   RETURN
END SUBROUTINE dv_release

END MODULE dynvect
