! ----------------------------------------------------------------------
! This is part of the 'Solve Singular BVP' (SOLVSING) package developed
! by W.Auzinger, O.Koch, P.Kofler, E.Weinmueller,
! Department of Applied Mathematics and Numerical Analysis,
! Vienna University of Technology, Vienna, Austria.
! ----------------------------------------------------------------------
MODULE ieulidec
! 'Implicit EULer IDEC method for singular ivps' (IEULIDEC)
! Version 1.09, last modified 05-05-2001.
! --- Function and Algorithm ---
! Integrate the IVP of ODE y'(x) = f(x,y) with constant step size using
! implicit Euler's method with global error estimation by Zadunaisky's
! technique and Iterated Defect Correction (IDeC). Check the global
! error and readjust the step size to satisfy a given global tolerance.
! Repeat the integration if is is necessary. (This tolerance checking
! may be skipped to examine asymtotic behaviour.)
! This is a general purpose module for integrating singular IVPs and
! avoiding order reduction. But away from the singularity, the used
! method is quite inefficient.
! --- Compiling ---
! This module has to be compiled third, following the linalg module and
! it is not dependant on function data. The usage of ivp_ieuleridec can
! be seen in t_idec_a.f90 for asymtotic usage (with tolerance checking
! deactivated) and in t_idec_t.f90 for global tolerance usage.
! --- Bibliography ---
! W.Auzinger,O.Koch,P.Kofler,E.Weinmueller,
! 'The Application of Shooting to Singular Boundary Value Problems',
! Technical Report No. 126, Austria 1999.
! M.Metcalf, J.Reid,
! 'Fortran 90 Explained', Oxford University Press (1990),
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996).
! P.E.Zadunaisky,
! 'On the Estimation of Errors Propagated in the Numerical Integration
! of Ordinary Differential Equations', Numer.Math. 27(1976), pp. 21--39.
! ----------------------------------------------------------------------
! Constants
! ----------------------------------------------------------------------
   USE TYPES                 ! predefined data-types
   IMPLICIT NONE
   REAL(KIND = dp), PARAMETER :: safety_factor_stepsize = 0.9_dp, & 
                    min_update_stepsize = 2.0_dp, &
                    max_update_stepsize = 8.0_dp, &
                    dumping_factor_newton = 1.0_dp, & 
                    default_min_stepsize = 10.0_dp*EPSILON(1.0_dp), &
                    relative_error_factor = 1.0_dp
!        safety_factor_stepsize ... safety factor for new step sizes.
!           If automatic step size control is enabled (stepcontrol=true)
!           the new step-size is multiplied by this factor to be on the
!           save side, as the new step-size estimation is not accurate.
!           But smaller factors make the integration inefficient. If the
!           step sizes tend to be too large for special problems, which
!           can be seen by more than two iterations over the given
!           intervall, this factor should be decreased. <= 1.0_dp
!        min_update_stepsize ... min update factor for new step sizes.
!           If automatic step size control is enabled (stepcontrol=true)
!           and the step size has to be reduced it is at least divided
!           by this factor. To deactivate this, use 0.0_dp. >= 0.0_dp
!        max_update_stepsize ... max update factor for new step sizes.
!           If automatic step size control is enabled (stepcontrol=true)
!           and the step size has to be reduced it is at most divided
!           by this factor to avoid too fast reduction. To deactivate
!           this, max_update_stepsize = HUGE(1.0_dp). > 0.0_dp
!        dumping_factor_newton ... factor first dumped Newton iteration
!           If damping for Newton iteration is enabled (damping=true),
!           the first damped step is the full step multiplied by this
!           factor. Then the factor is taken half in each further damped
!           step. <= 1.0_dp
!        default_min_stepsize ... default minimal stepsize
!           If the minimum step-size parameter is skipped, this is used.
!        relative_error_factor ... the factor the global error estimate
!           has to be less than the solution itself to indicate a valid
!           error estimation, or otherwise cause error 5. <= 1.0_dp
   INTEGER(KIND = i4b), PARAMETER :: max_newton_steps = 25, &
                                     max_lambda_steps = 10
!        max_newton_steps ... maximum of Newton iterations for Euler
!           If the Newton interation for nonlinear problems in the
!           Implicit-Euler method needs more than this number of steps,
!           most likely there is no convergence, thus the method stops,
!           reporting an error. For special problems with very slow
!           convergence, this number may be increased.
!        max_lambda_steps ... maximum of damping steps in Newton method
!           Newton convergence is improved by damping (damping=true),
!           using decreasing steps in the direction given by the
!           Newton increment. This is the number of damped steps to
!           try. If no satisfying solution is found, most likely there
!           is no convergence, thus the method stops reporting an error.
   LOGICAL, PARAMETER :: default_damping = .true., &
                         default_stepcontrol = .true.
!        default_damping ... default use damping for Newton iteration
!        default_stepcontrol ... default use of step-control

   PRIVATE
   PUBLIC :: ivp_ieuleridec

CONTAINS
! ----------------------------------------------------------------------
SUBROUTINE ivp_ieuleridec(fcn, dfcn, a, y, b, linear, h, global_tol, &
           global_rel, newton_tol, order_method, ifail, solution, &
           number_iter, min_stepsize, order_poly, use_damping, &
           use_stepcontrol)
! ----------------------------------------------------------------------
! Parameters: fcn(in), dfcn(in), a(in), y(in/out), b(in), linear(in),
! h(in/out), global_tol(in/out), global_rel(in/out), newton_tol(in),
! order_method(in), ifail(out), solution(opt out), number_iter(opt out),
! min_stepsize(opt in), order_poly(opt in), use_damping(opt in),
! use_stepcontrol(opt in)
! ----------------------------------------------------------------------
! required Arguments
! ----------------------------------------------------------------------
   USE LINALG, ONLY : lubksb, ludcmp ! we need the LU decomposition
   IMPLICIT NONE
   INTERFACE
      FUNCTION fcn(t, y)     ! right side of the IVP, y'(t) = fcn(t, y)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: fcn(SIZE(y))
      END FUNCTION fcn
      FUNCTION dfcn(t, y)    ! Jacobian of fcn(t, y)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: dfcn(SIZE(y), SIZE(y))
      END FUNCTION dfcn
   END INTERFACE
   REAL(KIND = dp), INTENT(in) :: a, b, newton_tol
!        a, b ... interval to perform the integration
!        newton_tol ... tolerance of Newton-iteration in implicit Euler
!           method. Both Newton-residual and increment are checked.
   REAL(KIND = dp), INTENT(in out) :: y(:), h, global_tol, global_rel
!        y ... (in) initial value y(a), (out) approximation for y(b)
!        h ... (in) starting step, (out) last used step, or suggestion
!           for better step if step control is disabled, use h<0 for tvp
!        global_tol ... (in) global tolerance to fulfill,
!           (out) estimation of the maximum of the global error on a..b
!        global_rel ... (in) global relative tolerance to fulfill,
!           (out) estimation of the relative global error on a..b
   LOGICAL, INTENT(in) :: linear
!        linear ... true if the initial value problem is linear
   INTEGER(KIND = i4b), INTENT(in) :: order_method
!        order_method ... order of the method = iterations of IDeC + 1
   INTEGER(KIND = i4b), INTENT(out) :: ifail
!        ifail ... error variable, set due to several errors
!            0 = everything is ok, integration was successfull
!           Negative numbers indicate wrong or unproper input values:
!           -1 = illegal interval range, error in parameters a,b
!           -2 = illegal smallest step size, error in min_stepsize
!           -3 = illegal starting step size, error in parameter h
!           -4 = illegal global tolerance, error in global_tol/rel
!           -5 = illegal Newton tolerance, error in newton_tol
!           -6 = illegal method, error in order_method, order_polynomial
!           Positive numbers indicate problems during the integration.
!           In these cases a solution *may* be available, but does not
!           fulfill given tolerance requirements.
!            1 = could not complete integration due to h < h_min
!            2 = Newton-iteration needed too many steps, indicating bad
!                or no convergence of Newton's method
!            3 = global error did not decrease during IDeC
!            4 = Newton residual did not decrease during the iteration,
!                indicating bad or no convergence of Newton's method
!            5 = the global error (estimate) was larger than 100 percent
!                indicating a step size which is much too large.
! ----------------------------------------------------------------------
! optional Arguments
! ----------------------------------------------------------------------
   REAL(KIND = dp), OPTIONAL, POINTER :: solution(:,:)
!        solution ... the estimation of the complete solution at the
!           used grid (y_0,y_1,...y_N), array indices starting at 0.
   INTEGER(KIND = i4b), OPTIONAL, INTENT(out) :: number_iter
!        number_iter ... number of iterations over stepsize, 1-4. If
!           this returns a value >> 2 on output, the safety factor may
!           be too large for the given problem.
   REAL(KIND = dp), INTENT(in), OPTIONAL :: min_stepsize
!        min_stepsize ... smallest absolute allowed step size, >= 0.0
   INTEGER(KIND = i4b), INTENT(in), OPTIONAL :: order_poly
!        order_poly ... order Zadunaisky-polynomials (>=idec_iter+1)
   LOGICAL, INTENT(in), OPTIONAL :: use_damping, use_stepcontrol
!        use_damping ... use damping for Newton iteration
!        use_stepcontrol ... use step-control to fullfill tolerance. If
!           not, this subroutine returns after the first integration.
! ----------------------------------------------------------------------
! Local variables for optional arguments or their defaults
! ----------------------------------------------------------------------
   REAL(KIND = dp) :: h_min
!        h_min ... used smallest step size from min_stepsize or default.
   INTEGER(KIND = i4b) :: pord, idec_iter
!        pord ... used order polynomials (>=idec_iter+1)
!        idec_iter ... number of iterations of IDeC
   LOGICAL :: damping, stepcontrol
!        damping ... use damping for Newton iteration
!        stepcontrol ... use step-control to fullfill tolerance
! ----------------------------------------------------------------------
! Local variables
! ----------------------------------------------------------------------
   INTEGER(KIND = i4b) :: j, iter, step, pi, feldi, idec, maxsteps
!        j ... loop counter
!        iter ... iteration counter for step size/tolerance iteration
!        step ... step counter in integration from a to b
!        pi ... step counter for little steps for Zadunaisky-polynomials
!        feldi ... reference value for interpolation in yfield
!        idec ... counter IDeC iteration
!        maxsteps ... total number of steps in the integration
   REAL(KIND = dp) :: ya(SIZE(y)), x, est(SIZE(y)), global_est, ratio, &
                      old_h, old_est, global_y, old_y, oop
!        ya ... store initial value for start of several integrations
!        x ... current position during the integration in [a,b]
!        est ... estimation of a single global error by Zadunaisky
!        global_est ... maximum of norm of estimations of global error
!        ratio ... step size-update ratio
!        old_h ... last used step size
!        old_est ... estimated error of solution of last iteration
!        global_y ... maximum of norm of estimations of solution
!        old_y ... maximum of norm of estimations of last iteration
!        oop ... exponent for ratio in error estimation ( 1/(order) )
   REAL(KIND = dp), ALLOCATABLE :: yfield(:,:),yfieldw(:,:),yfieldb(:,:)
!        yfield ... store all approximations to interpolate the solution
!        yfieldw ... approximations workspace to improve solution
!        yfieldb ... store basic approximations for IDeC
! ----------------------------------------------------------------------
! Code
! ----------------------------------------------------------------------
   IF (debug_output) WRITE(*,'(A)') '  ---> ivp_ieuleridec called'

! check the input arguments and set up the optional arguments
   IF (PRESENT(number_iter)) number_iter = 0
   ifail = -1                ! check interval
   IF ( ((h>0.0_dp).AND.(b<= a)) .OR. ((h<0.0_dp).AND.(a<=b)) ) RETURN
   ifail = -2                ! check smallest step size
   IF (PRESENT(min_stepsize)) THEN
      h_min = min_stepsize   ! min stepsize was supplied
   ELSE
      h_min = default_min_stepsize ! use default
   END IF
   IF ( (h_min < 0.0_dp)  .OR. & 
      (h_min <= EPSILON(1.0_dp)) ) RETURN ! too small
   IF (h_min > ABS(b-a)) RETURN ! too large
   ifail = -3                ! check starting step size
   IF (h == 0.0_dp) RETURN
   IF (ABS(h) < h_min) RETURN ! too small, or use h = SIGN(h_min,h)
   IF (ABS(h) > ABS(b-a)) RETURN ! too large, or use h = SIGN(b-a,h)
   ifail = -4                ! check global tolerance
   IF (global_tol <= 0.0_dp) RETURN
   IF ( (global_rel < 0.0_dp) .OR. (global_rel > 1.0_dp) ) RETURN
   IF (.not. linear) THEN
      ifail = -5             ! check Newton-tolerance
      IF (newton_tol <= 0.0_dp) RETURN
      ! IF (newton_tol > global_tol) RETURN ! should at least be exact
                             ! but this should rather be a warning, not
                             ! an error, thus we do not check it
   END IF
   ifail = -6                ! check order and number of iterations
   IF (order_method < 1) RETURN
   idec_iter = order_method - 1
   oop = 1.0_dp              ! the order estimation is for order-1
   IF (order_method > 1) oop = 1.0_dp/(order_method-1)
   IF (PRESENT(order_poly)) THEN
      pord = order_poly      ! order was supplied
      ! IF (pord < order_method) RETURN ! should not be, maybe for tests
   ELSE
      pord = order_method    ! use default
   END IF
   IF (PRESENT(use_damping)) THEN
      damping = use_damping
   ELSE
      damping = default_damping ! use default
   END IF
   IF (PRESENT(use_stepcontrol)) THEN
      stepcontrol = use_stepcontrol
   ELSE
      stepcontrol = default_stepcontrol ! use default
   END IF
   IF ((stepcontrol).AND.(order_method<2)) RETURN ! we do not have an
                             ! estimation for order 1, as we need at
                             ! least 1 IDeC iteration, yielding order 2

! verbose the used parameters to the debug output if wished
   IF (debug_output) THEN
      WRITE(*,*) ' * ivp: ',a,' to ',b,' (linear',linear,')'
      WRITE(*,'(A,I1,A,I1,A,I1,A)') '  * method: order ',order_method, &
             ' (polynomial ',pord,' with ',idec_iter,' IDeC iterations)'
      WRITE(*,'(A,E14.8,A,E14.8,A)') '  * stepsize: initial ',h, &
             ' (minimal ',h_min,')'
      WRITE(*,*) ' * control: global tolerance ', global_tol, &
             ' (step size control',stepcontrol,')'
      WRITE(*,*) ' * Newton: tolerance ', newton_tol, ' (with damping', &
             damping,')'
   END IF ! debug_output

! set up internal variables
   ifail = 0
   iter = 0
   ya = y                    ! store initial value y(a)
   old_est = HUGE(1.0_dp)
   old_y = 1.0_dp

! do the big step size/tolerance iteration until the maximum of the
! estimated global error is below the given tolerance
   DO                        ! big iteration over step size/tolerance
      iter = iter + 1

! calculate amount of necessary steps and adjust h to distribute grid
! points equally spaced on [a,b], where the number of steps must be a
! multiple of pord. h itselt is the real step size, thus h*pord is the
! desired size of the big steps used by Zadunaisky's technique.
      maxsteps = CEILING((b-a)/(h*pord)/(1+EPSILON(1.0_dp)))
                             ! use h*(1+EPSILON) to avoid a further step
                             ! due to round-off error in h
      IF (maxsteps < 1) maxsteps = 1 ! one big step
      h = (b-a)/maxsteps/pord ! in any case we have hnew <= h
      old_h = h              ! save old stepsize
      IF (debug_output) WRITE(*,'(A,I1,A,I5,A,E14.8)') '  Iteration: ',&
         iter,' performing ', maxsteps,' steps with step size ',h

! now check new adjusted step size again towards parameters
      IF (ABS(h) < h_min) THEN
         IF (iter == 1) THEN ! we did not do anything yet
            ifail = -3       ! initial step size was too small
         ELSE                ! iter > 1, so a solution is available
            ifail = 1        ! could not complete because of h < h_min
            global_tol = old_est ! that is what we got in last iteration
            IF (old_y /= 0.0_dp) THEN
               global_rel = old_est/old_y ! relative global error
            ELSE
               global_rel = old_est
            END IF
            h = old_h        ! with the last used step size
                             ! solution has the values of the last run
           iter = iter -1    ! the last iteration failed
         END IF
         EXIT                ! skip all further iterations
      END IF ! ABS(h) < h_min

! drop the last solution and inialize new soultion field at base of heap
      IF (PRESENT(solution)) THEN
         IF (ASSOCIATED(solution)) DEALLOCATE(solution) ! drop last
         ALLOCATE(solution(0:maxsteps*pord,SIZE(y)))
      END IF
! allocate workspace after solution, thus must be freed before
      ALLOCATE(yfield(0:maxsteps*pord,SIZE(y)),yfieldw(0:pord,SIZE(y)),&
               yfieldb(0:maxsteps*pord, SIZE(y)))

! calculate the basic approximation using implicit Euler's method
      x = a
      step = 0
      y = ya                 ! initialise initial values
      yfield(step,:) = y     ! store first value, the initial value
      global_y = MAXVAL(ABS(y)) ! norm of y(a), set for idec_iter=0
      global_est = 0.0_dp ! y(a) is exact, thus error is 0
      DO j = 1, maxsteps     ! loop for big steps
         DO pi = 1, pord     ! loop for little, real steps in polynomial
            IF (linear) THEN
               y = impl_lin_step(h,.false.)
            ELSE
               IF (damping) THEN
                  y = impl_nlin_damp_step(h,.false.)
               ELSE
                  y = impl_nlin_step(h,.false.)
               END IF
               IF (ifail /= 0) THEN ! stop integration due to error
                  CALL error_in_loop()
                  RETURN
               END IF ! ifail /= 0
            END IF
            x = x + h        ! go to next grid point
            step = step + 1
            yfield(step,:) = y ! store basic approximation
         END DO ! DO pi over little steps
      END DO ! DO j over steps
      yfieldb = yfield       ! store the basic approximation of order 1

! start IDeC iteration, we do at least one iteration, as is order >= 2
      DO idec = 1, idec_iter

! calculate idec-th approximations for the neighbouring problem
         x = a               ! initialise initial values
         step = 0
         y = yfield(step,:)
         yfieldw(0,:) = y
         global_y = MAXVAL(ABS(y)) ! norm of y(a)
         global_est = 0.0_dp ! y(a) is exact, thus error is 0
         DO j = 1, maxsteps  ! loop for big Zadunaisky steps
            feldi = step
            DO pi = 1, pord  ! loop for little, real steps in polynomial
               IF (linear) THEN
                  y = impl_lin_step(h,.true.)
               ELSE
                  IF (damping) THEN
                     y = impl_nlin_damp_step(h,.true.)
                  ELSE
                     y = impl_nlin_step(h,.true.)
                  END IF
                  IF (ifail /= 0) THEN ! stop integration due to error
                     CALL error_in_loop()
                     RETURN
                  END IF ! ifail /= 0
               END IF
               x = x + h     ! go to next grid-points
               step = step + 1
               est = y-yfieldb(step,:) ! error estimation for this y
               yfieldw(pi,:) = yfield(step,:) - est ! update solution
               global_y = MAX(global_y,MAXVAL(ABS(y))) ! get maximum
               global_est = MAX(global_est,MAXVAL(ABS(est))) ! get max
            END DO ! DO pi over little steps
            yfield(step-pord:step-1,:) = yfieldw(0:pord-1,:) 
            yfieldw(0,:) = yfieldw(pord,:) ! last point becomes first
         END DO ! DO j over steps
         yfield(step,:) = yfieldw(pord,:)

         ! alternatively we could use the estimation at b,
         ! global_est = MAXVAL(ABS(y-yfieldb(step,:)))
         ! but we are interested in BVPs, thus we take the maximas

         IF (debug_output) WRITE(*,'(A,I1,A,I1,A,E14.8)') &
            '  Iteration: ', iter, ' in IDeC iteration ', idec, &
            ' max.error est. ', global_est

         IF ( (stepcontrol) .AND. &
              (global_est < global_tol + global_y * global_rel) ) EXIT
         ! solution is accurate before we used all our idec iterations
      END DO ! DO idec over iterations

! store the last, most accurate approximation, which is is of order pord
      y = yfield(step,:)     ! y approximates y(b)
      IF (PRESENT(solution)) solution = yfield ! the last whole solution
      DEALLOCATE(yfield,yfieldw,yfieldb) ! freeing data leaving solution

! check if step size was small enough to produce a correct error
! estimation, i.e. if the error estimation is smaller than the solution,
! global error/solution < 100 percent
      IF (global_est > relative_error_factor*global_y) THEN
         ifail = 5           ! global error is 100 percent
         global_tol = global_est ! we got this with last step size
         IF (global_y /= 0.0_dp) THEN
            global_rel = global_est/global_y ! relative error
         ELSE
            global_rel = global_est
         END IF
         EXIT                ! we also have a solution, but useless
      END IF ! global_est > global_y
      IF (debug_output) WRITE(*,'(A,I1,A,E14.8,A,E14.8)') &
         '  Iteration: ', iter, ' error estimation ', global_est, &
         ' local error est. ', global_est/global_y

! check the global error, accept solution or refine the step size
      IF ( (global_est <= global_tol + global_y * global_rel) .OR. &
           (.not. stepcontrol) ) THEN
         global_tol = global_est ! save estimation to return it
         IF (global_y /= 0.0_dp) THEN
            global_rel = global_est/global_y ! relative error
         ELSE
            global_rel = global_est
         END IF
         EXIT                ! OK, we succeeded and have a solution
                             ! without step control do just 1 iteration
      ELSE                   ! reject solution, calculate new step size
         IF (global_est >= old_est) THEN
            ifail = 3        ! error is not decreasing
            global_tol = global_est ! we got this with last step size
            IF (global_y /= 0.0_dp) THEN
               global_rel = global_est/global_y ! relative error
            ELSE
               global_rel = global_est
            END IF
                             ! we also have a solution
            EXIT
         END IF
         old_h = h           ! store old h and old error estimation
         old_est = global_est
         old_y = global_y

! calculate the new stepsize upon our successfull error estimation
         ratio = (global_est / (global_tol+global_y*global_rel) )**oop
         ! the order of method is order_method, but the global error
         ! estimation is for order_method-1, as derived by the last
         ! iteration, which is of course used for the solution.
         IF (ratio < min_update_stepsize) ratio = min_update_stepsize
         IF (ratio > max_update_stepsize) ratio = max_update_stepsize
         h = h/ratio*safety_factor_stepsize ! reduced step size
         IF (debug_output) WRITE(*,'(A,I1,A,E14.8,A,E14.8)') &
            '  Iteration: ', iter, ' update ratio ', ratio, &
            ' yields new step size ', h
      END IF ! global_est <= global_tol
   END DO ! iter
   IF (PRESENT(number_iter)) number_iter = iter
   IF (debug_output) WRITE(*,'(A)') '  <--- ivp_ieuleridec returned'

   RETURN
! ----------------------------------------------------------------------
! Local subroutines
! ----------------------------------------------------------------------
   CONTAINS

SUBROUTINE error_in_loop()
! --- Function and Algorithm ---
! Perform cleanup for errors during the do-loops, i.e. errors in the
! implicit Euler method.
! --- Arguments ---
   IMPLICIT NONE
! --- Code ---
   DEALLOCATE(yfield,yfieldw,yfieldb) ! freeing data
   IF (PRESENT(solution)) DEALLOCATE(solution)
   IF (iter > 1) THEN        ! an old solution is available,
                             ! but we have no solution vector for that
      global_tol = old_est   ! we got in last iteration
      IF (old_y /= 0.0_dp) THEN
         global_rel = old_est/old_y ! relative error
      ELSE
         global_rel = old_est
      END IF
      h = old_h              ! with the last used step size
      iter = iter -1         ! the last iteration failed
      IF (PRESENT(number_iter)) number_iter = iter
   END IF
   RETURN
END SUBROUTINE error_in_loop

FUNCTION lagrang()
! Version 1.01, last modified 07-09-1998.
! --- Function and Algorithm ---
! Evaluate the polynomial sum_{j=0}^pord omega_j*f_j (lagrange-poly)
! interpolating yfield on [a,a+pord*h] at a given x. As x consists just
! of grid points for Implicit Euler,just return the value of
! yfield(step+1).
! required temporary storage: SIZE(y)
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: lagrang(SIZE(y))
!   REAL(KIND = dp), INTENT(in) :: x
!        x ... posititon the evaluation of lagrange-poly is required
! --- Global variables ---
!        step, yfield(:,:), y (size only)
! --- Code ---
   lagrang = yfield(step+1,:)
   RETURN
END FUNCTION lagrang

FUNCTION lagran2(x)
! Version 1.00, last modified 14-01-1998.
! --- Function and Algorithm ---
! Evaluate the polynomial sum_{j=0}^pod omega_j*f_j (lagrange-poly)
! interpolating yfield on [a,a+pord*h] at x. This is the general
! purpose method, which is not used by the main method.
! required temporary storage: 2*SIZE(y)+5
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: lagran2(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: x
!        x ... posititon the evaluation of lagrange-poly is required
! --- Global variables ---
!        yfield(:,:), feldi, pord, h, y (size only)
! --- Local variables ---
   REAL(KIND = dp) :: u, w, ph(SIZE(y))
!        u ... x-value shifted to [0, pord*h]
!        w ... auxiliary data
!        ph ... value of lagrange-polynomial at x
   INTEGER(KIND = i4b) :: j, k
!        j, k ... loop counters
! --- Code ---
   u = x - a - h*feldi       ! move interval [a+...,...] to [0,...]
   ph = 0.0_dp               ! initialise sum of omega_j*f_j
   DO j = 0, pord            ! calculate sum omega_j*f_j
      w = 1.0_dp
      DO k = 0, pord ! calculate product omega_j (ord_poly fctrs)
         IF (j /= k) w = w * (u-h*k)/(j-k)
      END DO
      ph = ph + w*yfield(feldi+j,:) ! add omega_j*f_j
   END DO
   lagran2 = ph/h**pord
   RETURN
END FUNCTION lagran2

FUNCTION difflagr(x)
! Version 1.01, last modified 07-09-1998.
! --- Function and Algorithm ---
! Evaluate the derivative of the polynomial sum_{j=0}^pord omega_j*f_j
! (lagrange-polynomial) interpolating yfield on [a,a+pord*h] at x=k*h.
! required temporary storage: 2*SIZE(y)+7
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: difflagr(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: x
!        x ... value the evaluation is required
! --- Global variables ---
!        a, h, feldi, pord, yfield(:,:), y(size only)
! --- Local variables ---
   REAL(KIND = dp) :: w, w2, phs(SIZE(y))
!        u ... x-value shifted to [0,pord*h]
!        w, w2 ... auxiliary data
!        phs ... value of (lagrange-polynomial)' at x
   INTEGER(KIND = i4b) :: i, j, k, l
!        i ... index of evaluation point in yfield
!        j, k, l, u ... loop counter
! --- Code ---
   i = nint((x-a-h*feldi)/h) ! move interval to [0,...]
   phs = 0.0_dp              ! initialise sum of omega'_j*f_j

   DO j = 0, pord            ! calculate sum omega'_j*f_j(pord+1 sum.)
      w = 0.0_dp             ! initialise sum for omega'_j
      DO k = 0, pord         ! calculate sum omega'_j (pord sumands)
         IF ((j /= k).and.((i == j).or.(i == k))) THEN
            w2 = 1.0_dp      ! initialise k-th product in omega'_j
            DO l = 0, pord   ! calculate k-th product in omega'_j
               IF ((j /= l).and.(l /= k)) w2 = w2*(i-l)
            END DO
            w = w + w2       ! add components of omega'_j
         END IF
      END DO

      DO k = 0, pord         ! calculate the constant factor of omega_j
         IF (j /= k ) w = w/(j-k)
      END DO
      phs = phs + w*yfield(feldi+j,:) ! add omega'_j*f_j
   END DO
   difflagr = phs/h
   RETURN
END FUNCTION difflagr

FUNCTION expl_lin_step(h, zadu)
! Version 1.00, last modified 07-09-1998.
! --- Function and Algorithm ---
! Perform one explicit Euler step for fcn or neighbouring problem.
! required temporary storage: 2*SIZE(y)+2
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: expl_lin_step(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: h
!        h ... step size to perform implicit Euler step
   LOGICAL, INTENT(in) :: zadu
!        zadu ... use Zadunaisky polynomials for neighbouring problem
! --- Global variables ---
!        fcn, x, y
!        for zadu indirect usage of yfield(:,:), feldi, pord
! --- Local variables ---
   REAL(KIND = dp) :: w(SIZE(y)), lag(SIZE(y))
!        w ... inhomogeous part of fcn
!        lag ... explicit evaluation of lagran2 for IBM AIX XL Fortran
! --- Called software components ---
!        direct ... for zadu difflag2
! --- Code ---
   w = y + h*fcn(x,y)
   IF (zadu) THEN
!     w = w + h*(difflagr(x) - fcn(x, lagrang(x)))
      lag = lagrang()
      lag = difflagr(x) - fcn(x, lag)
      w = w + h*lag
   END IF
   expl_lin_step = w         ! new solution for y
   RETURN
END FUNCTION expl_lin_step

FUNCTION impl_lin_step(h, zadu)
! Version 1.01, last modified 03-09-1998.
! --- Function and Algorithm ---
! Perform one implicit Euler step for linear or neighbouring problem.
! required temporary storage: SIZE(y)^2+4*SIZE(y)+4
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: impl_lin_step(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: h
!        h ... step size to perform implicit Euler step
   LOGICAL, INTENT(in) :: zadu
!        zadu ... use Zadunaisky polynomials for neighbouring problem
! --- Global variables ---
!        fcn, dfcn, x, y
!        for zadu indirect usage of yfield(:,:), feldi, pord, h
! --- Local variables ---
   REAL(KIND = dp) :: Jac(SIZE(y),SIZE(y)), w(SIZE(y)), &
                      u, lag(SIZE(y))
!        Jac ... Jacobian for given x
!        w ... inhomogenous part of fcn
!        u ... auxiliary data
!        lag ... explicit evaluation of lagrang for IBM AIX XL Fortran
   INTEGER(KIND = i4b) :: indx(SIZE(y)), i
!        indx ... pivoting indices for LU-decomposition
!        i ... loop counter
! --- Called software components ---
!        direct ... ludcmp, lubksb, for zadu difflagr
! --- Code ---
! set up linear system Jac*y1=w for implicit Euler
   w = 0.0_dp
   u = x + h
   w = y + h*fcn(u,w)        ! inhomogenous component of problem
   IF (zadu) THEN
!     w = w + h*(difflagr(u) - fcn(u, lagrang(u)))
      lag = lagrang()
      lag = difflagr(u) - fcn(u, lag)
      w = w + h*lag
   END IF
   Jac = -h*dfcn(u,y)        ! homogenous component of problem
                             ! y dummy argument because linear
   DO i = 1, SIZE(y)         ! loop over diagonal, add identity
      Jac(i,i) = Jac(i,i) + 1.0_dp
   END DO ! DO i over diagonal

! solve linear system Jac*y1=w, destroying Jac and yielding y1 in w
   CALL ludcmp(Jac, indx)
   CALL lubksb(Jac, indx, w)

   impl_lin_step = w         ! y1 = new solution for y
   RETURN
END FUNCTION impl_lin_step

FUNCTION impl_nlin_step(h, zadu)
! Version 1.01, last modified 03-09-1998.
! --- Function and Algorithm ---
! Perform one Implicit Euler step for non-linear fcn or the neighbouring
! problem using simple Newton iteration.
! required temporary storage: SIZE(y)^2+5*SIZE(y)+7
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: impl_nlin_step(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: h
!        h ... step size to perform implicit Euler step
   LOGICAL, INTENT(in) :: zadu
!        zadu ... use Zadunaisky polynomials for neighbouring problem
! --- Global variables ---
!        fcn, dfcn, x, y, Newton_tol, constant max_newton_steps
!        for zadu indirect usage of yfield(:,:), feldi, pord, h
! --- Local variables ---
   REAL(KIND = dp) :: Jac(SIZE(y),SIZE(y)), w(SIZE(y)), &
                      y1(SIZE(y)), u, nrmF, nrmdeltak, lag(SIZE(y))
!        Jac ... Jacobi-matrix for given x
!        w ... fcn call
!        nrmF ... norm of residual
!        nrmdeltak ... norm of Newton-increment (deltak)
!        y1 ... approximation of step
!        u ... auxiliary data
!        lag ... explicit evaluation of lagran2 for IBM AIX XL Fortran
   INTEGER(KIND = i4b) :: indx(SIZE(y)), i, index
!        indx ... pivoting indices for LU-decomposition
!        i ... loop counter
!        index ... iteration count in Newton-iteration
! --- Called software components ---
!        direct ... ludcmp, lubksb, for zadu difflagr
! --- Code ---
! get initial approximation for y1
   y1 = y
   u = x + h

! start Newton-iteration
   index = 0
   DO
      index = index + 1

! set up linear system Jac*y1^(i) = w for Newton-increment
      w = y + h*fcn(u, y1) - y1 ! calculate Newton-residual
      IF (zadu) THEN
!        w = w + h*(difflagr(u) - fcn(u, lagrang(u)))
         lag = lagrang()
         lag = difflagr(u) - fcn(u, lag)
         w = w + h*lag
      END IF
      nrmf = MAXVAL(ABS(w))  ! Newton-residual for this approximation
      Jac = h*dfcn(u, y1)    ! Jacobi-matrix
      DO i = 1, SIZE(y)      ! loop over diagonal, subtract identity
         Jac(i,i) = Jac(i,i) - 1.0_dp
      END DO ! DO i over diagonal

! solve linear system, destroying Jac and yielding delta y1 in w
      CALL ludcmp(Jac, indx)
      CALL lubksb(Jac, indx, w)

! iterate solution and exit if accurate
      y1 = y1 - w
      nrmdeltak = MAXVAL(ABS(w)) ! size of Newton-increment
      IF ((nrmf < Newton_tol) .AND. (nrmdeltak < Newton_tol)) EXIT

      IF (index > max_newton_steps) THEN
         ifail = 2           ! error, too many iterations
         impl_nlin_step = 0.0_dp
         RETURN
      END IF
   END DO ! Newton-loop

   impl_nlin_step = y1       ! iterated y1 = new solution for y
   RETURN
END FUNCTION impl_nlin_step

FUNCTION impl_nlin_damp_step(h, zadu)
! Version 1.03, last modified 18-11-1998.
! --- Function and Algorithm ---
! Perform one Implicit Euler step for non-linear fcn or the neighbouring
! problem using the damped Newton's method.
! required temporary storage: SIZE(y)^2+8*SIZE(y)+8
! --- Bibliography ---
! Deuflhard,
! 'Numerische Mathematik: Eine algorithmisch orientierte Einfhrung',
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: impl_nlin_damp_step(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: h
!        h ... step size to perform implicit Euler step
   LOGICAL, INTENT(in) :: zadu
!        zadu ... use Zadunaisky polynomials for neighbouring problem
! --- Global variables ---
!        fcn, dfcn, x, y, newton_tol, constant max_newton_steps
!        for zadu indirect usage of yfield(:,:), feldi, pord, h
! --- Local variables ---
   REAL(KIND = dp) :: Jac(SIZE(y),SIZE(y)), res(SIZE(y)), lambda, &
                      delta(SIZE(y)), y1(SIZE(y)), y2(SIZE(y)), u, &
                      lag(SIZE(y)), nrmdeltak, w(SIZE(y))
!        Jac ... Jacobian for given x
!        res ... Newton-residual
!        delta ... Newton-increment
!        nrmdeltak ... norm of Newton-increment (delta)
!        y1 ... approximation of step
!        y2 ... approximation of step used by damping
!        u,w ... auxiliary data, work
!        lag ... explicit evaluation of lagrang
!        lambda ... last damping factor
   INTEGER(KIND = i4b) :: indx(SIZE(y)), i, index, lindex
!        indx ... pivoting indices for LU-decomposition
!        i ... loop counter
!        index ... iteration count in Newton-iteration
!        lindex ... iteration count in damping loop
! --- Called software components ---
!        direct ... ludcmp, lubksb, for zadu difflagr
! --- Code ---
   y1 = y                    ! initial approximation for y1
   u = x + h                 ! target point
   lambda = dumping_factor_newton/2.0_dp ! initial damping factor

! calculate 1st Newton-residual
   res = y+h*fcn(u, y1) - y1 ! calculate Newton-residual
   IF (zadu) THEN
!     res = res + h*(difflagr(u) - fcn(u, lagrang(u)))
      lag = lagrang()
      lag = difflagr(u) - fcn(u, lag)
      res = res + h*lag
   END IF

! start Newton-iteration
   index = 0
   DO
      index = index + 1

! set up linear system Jac*y1^(i) = delta for Newton-increment
      delta = res
      Jac = h*dfcn(u, y1)    ! Jacobian
      DO i = 1, SIZE(y)      ! loop over diagonal, subtract identity
         Jac(i,i) = Jac(i,i) - 1.0_dp
      END DO ! DO i over diagonal

! solve linear system, destroying Jac and yielding delta_y1 in delta
      CALL ludcmp(Jac, indx)
      CALL lubksb(Jac, indx, delta)

! exit iteration if accurate
      nrmdeltak = MAXVAL(ABS(delta)) ! size of Newton-increment
      IF ((MAXVAL(ABS(res)) < newton_tol) .AND. &
          (nrmdeltak < newton_tol*(1.0_dp+MAXVAL(ABS(y1))))) EXIT

! update solution and iterate damping factors
      lambda = MIN(1.0_dp,lambda*2.0_dp)
      lindex = 0
      DO
         lindex = lindex + 1
         y2 = y1 - lambda*delta

! calculate new Newton-residual
         res = y+h*fcn(u, y2) - y2 ! calculate Newton-residual
         IF (zadu) res = res + h*lag

! check "monotonic test" of the residuals by the increments
         w = res             ! simplified Newton-increment
         CALL lubksb(Jac, indx, w)

         IF (MAXVAL(ABS(w)) <= (1.0_dp-lambda/2.0_dp)*nrmdeltak) THEN
            y1 = y2          ! "monotonic test" is true
            EXIT             ! therefore take this solution
         END IF

! shorten increment's length by 2
         lambda = lambda/2.0_dp
         IF (lindex > max_lambda_steps) THEN
            ifail = 4        ! error, residual did not decrease
            impl_nlin_damp_step = 0.0_dp
            RETURN
         END IF
      END DO ! lambda dumping
      IF (index > max_newton_steps) THEN
         ifail = 2           ! error, too many iterations
         impl_nlin_damp_step = 0.0_dp
         RETURN
      END IF
   END DO ! Newton-loop

   impl_nlin_damp_step = y1       ! iterated y1 = new solution for y
   RETURN
END FUNCTION impl_nlin_damp_step

! ----------------------------------------------------------------------
! End of ivp_ieuleridec
! ----------------------------------------------------------------------
END SUBROUTINE ivp_ieuleridec

END MODULE ieulidec
